//  -*- mode: c++; -*-
// Copyright (C) 2024
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef CA_H
#define CA_H

#include <vector>
#include <string>
#include <cstdint>
#include <iostream>
#include <boost/signals2.hpp>

#include "automaton.h"

// -- Namespaces: ---------------------------------------------------------
namespace P4 {

  // -- Type aliases: -------------------------------------------------------
  class CellularAutomaton;
  using CellularAutomatonPtr = CellularAutomaton*;
  using Cells_t = std::vector<uint8_t>;
  using nextgen_sig_t = boost::signals2::signal<void (uint32_t, const Cells_t&, const CellularAutomaton&)>;

  // -- Classes: ------------------------------------------------------------
  /// Esta clase representa automátas celulares unidimensionales.
  class CellularAutomaton : public Automaton {
  public:
    // -- Constructors/Destructor: --------------------------------------------
    /// Esta clase no tiene constructor por defecto
    CellularAutomaton() = delete;
    /// Constructor de copia
    CellularAutomaton(const CellularAutomaton &);
    /// Constructor a partir de ncells celdas, el carácter de dibujo c
    /// y la regla reset. En el estado inicial de todas las celdas del
    /// autómata solo la central (ncells/2) tendrá el valor 1, el
    /// resto valdrá 0.
    CellularAutomaton(uint64_t ncells, char c, uint8_t rset);
    /// Imprime en la salida estándar que el autómata ha sido
    /// destruido, tal y como puedes ver en la salida de ejemplo
    /// proporcionada en el enunciado.
    ~CellularAutomaton();

    // -- Setters: ------------------------------------------------------------
    /// Cambia la regla empleada en la evolución de este autómata
    void rule_set(uint8_t rset);

    // -- Output: -------------------------------------------------------------
    /// Redefine el operador de salida para mostrar en el stream os el
    /// número de celdas, la regla empleada en la evolucion y el
    /// carácter de dibujo del autómata ca tal y como puedes ver en la
    /// salida de ejemplo proporcionada en el enunciado.
    friend std::ostream &operator<<(std::ostream &os,
                                    const CellularAutomaton& ca);

    // -- Getters: ------------------------------------------------------------
    /// Devuelve como una cadena de 0s y 1s la representacion en
    /// binario del valor rset actual
    std::string get_ruleset() const;
    /// Devuelve un vector de 0s y 1s que representa la generacion
    /// actual de este automata
    const Cells_t &get_current_generation() const;
    /// Devuelve una cadena que representa la generacion actual. Cada
    /// carácter de esta cadena será o un espacio en blanco o el
    /// carácter de dibujo de este autómata dependiendo de si en el
    /// vector de 0s y 1s que representa la generación actual hay un 0
    /// o un 1 en la posición correspondiente
    std::string current_generation() const;

    // -- Action: -------------------------------------------------------------
    /// Imprime en el stream os la representacion como como cadena de
    /// la generacion actual. Esta cadena es la que devuelve
    /// CellularAutomaton::current_generation
    void print_current_generation(std::ostream &os) const;
    /// Calcula ngens generaciones de evolución para este automata,
    /// imprimiendo cada una de ellas en el stream os con un salto de
    /// linea al final y emitiendo la señal nxgen_sig despues de
    /// calcular cada nueva generacion
    void compute_ngens(uint32_t ngens, std::ostream& os);

    // -- Signals: ------------------------------------------------------------
    /// Señal emitida tras el cálculo de cada nueva generacion. Esta
    /// señal tiene como argumentos la generación actual, el vector de
    /// 0s y 1s que representa la generación actual y el autómata
    /// celular que la emite
    nextgen_sig_t nxgen_sig;

  private:
    /// Vector de 0s y 1s que tiene el resultado de la generación
    /// actual
    Cells_t cells;
    /// Cadena de 0s y 1s que contiene la representacion en binario
    /// del valor rset actual
    std::string ruleset;
  };
}

#endif /* CA_H */
