// -*- mode: c++ -*-
// Copyright (C) 2024
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef CONTAINER_H
#define CONTAINER_H

#include <vector>
#include <boost/signals2.hpp>
#include <string>

// -- Type aliases: -------------------------------------------------------
using namechange_sig_t = boost::signals2::signal<void (const std::string& oldname, const std::string& newname)>;

// -- Classes: ------------------------------------------------------------
namespace P4 {
  /// Plantilla de clase que representa la clase base de contenedores
  /// de datos de tipo T
  template <typename T>
  class Container {
  public:
    // -- Constructors/Destructor: --------------------------------------------
    /// Crea un Container con nombre name
    Container(const std::string &name);
    /// Implementación por defecto por si más adelante lo necesitamos redefinir
    virtual ~Container () = default;

    // -- Setters & Getters: --------------------------------------------------
    /// Permite cambiar el nombre a este contenedor por name
    void set_name(const std::string &name);
    /// Proporciona el nombre de este contenedor
    const std::string &get_name() const;
    /// Hace que todos elementos de este contenedor valgan d
    void set(const T &d);
    /// Añade un nuevo elemento d al final de este contenedor
    void add(const T &d);
    /// Devuelve el elemento i-esimo del contenedor
    T &operator[](size_t i);
    /// Devuelve el número de elementos en este contenedor
    size_t size() const;

    // -- Signals: ------------------------------------------------------------
    /// Señal emitida cuando se cambia el nombre a este contenedor
    namechange_sig_t namechange_sig;

  protected:
    /// Nombre del contenedor
    std::string cname;
    /// Almacen de los elementos de este contenedor
    std::vector<T> v;
  };
} // namespace P4

#endif /* CONTAINER_H */
